
# Created by Amin Rahmani, Ph.D., P.E. - February 3, 2018
# Copyright © 2018 Amin Rahmani. All rights reserved.

#========================================================================================
#                               ****   Notes  ****
# This code performs free-field site-response analysis in saturated soil condition.
# You need to makes changes on the lines that the comment "User-Defined" appears. Please go through all the command lines before running it.
# Units are in Metric. Check the units if you work with Imperial units. 
# The input motions should be available in two seperate .txt files. Time in "SCTime.txt" and Displacement in "SCDisp.txt".
# Outputs will be saved in a folder titled "Outputs".
#========================================================================================

wipe all

set startT [clock seconds]

################################################################
#### Input parameters used fro creating a stack of soil elements 
################################################################

set gravX  0.0000
set gravY  0.0000
set gravZ -9.8100				;# gravity in direction Z. The value would be -32.2 if Imperial units are used.

set numXele  1       
set numYele  1     
set numZele  13  				;# User-Defined - Number of elements in direction Z 

set numXnode  [expr $numXele+1]
set numYnode  [expr $numYele+1]
set numZnode  [expr $numZele+1]

set xSize   1.0                 ;# element size in direction x
set ySize   1.0                 ;# element size in direction y

#### You can apply 3 levels of mesh sizes in the vertical direction, z0, z1, and z2 represent layer thichness at the top, middle, and bottom layers.
set z0 0.70   					;# User-Defined - total thickness of zone z0
set z1 6.4 						;# User-Defined - total thickness of zone z1
set z2 4.8 						;# User-Defined - total thickness of zone z2
 
set zSize0    0.35              ;# User-Defined - element thickness in zone z0
set zSize1    0.8               ;# User-Defined - element thickness in zone z1
set zSize2    1.6               ;# User-Defined - element thickness in zone z2


###########################################################
#### Creating Nodes for 3-dimensional single column of soil
###########################################################


model basic -ndm 3 -ndf 4

set m 0
for {set i 1} {$i <= $numXnode} {incr i 1} {
  for {set j 1} {$j <= $numYnode} {incr j 1} {
    for {set k 1} {$k <= $numZnode} {incr k 1} {
       set xdim [expr ($i-1)*$xSize]
       set ydim [expr ($j-1)*$ySize]

      if { $k <= [expr ($z2/$zSize2+1)] } {
          set   zdim [expr ($k-1)*$zSize2]
      } else {
         if { $k > [expr ($z2/$zSize2)+($z1/$zSize1)+1] } {
          set   zdim [expr $z2+$z1+($k-(($z2/$zSize2)+($z1/$zSize1)+1))*$zSize0]
      } else {
          set   zdim [expr $z2+($k-(($z2/$zSize2)+1))*$zSize1]
      }
      }

       set   nodeNum   [expr $m*$numZnode+$k] 
       node  $nodeNum  $xdim $ydim $zdim
       puts " $nodeNum     $xdim     $ydim    $zdim"
    }
    set m [expr ($m+1)]
  } 
}


##########################################
#### Define Soil Materials
##########################################

# User-Defined --- define behavior of soil at different layers
### Mat 2
set Gs2            100000.0
set nou2     0.3
set Bs2      [expr 2.0*$Gs2*(1+$nou2)/3.0/(1-2.0*$nou2)]
set rhos2   [expr 126.5/32.2]
set cohes2  5.0
set fricAng2 33.0
set peakShearStra 0.10
set refPress 2089.0
set pressDependCoe 0.50
set PTAng2 25.5
set porosity2 0.424
set Perm2 [expr 5.953e-4/1.938/32.2]
set alpha2 [expr (pow($xSize,2))/(4.0*($Bs2 + (4.0/3.0)*$Gs2))]

nDMaterial PressureDependMultiYield02 2 3 $rhos2  $Gs2 $Bs2 $fricAng2 $peakShearStra $refPress $pressDependCoe $PTAng2 0.05 0.25 0.06 0.00 20 4.0 3.0 1.0 0.0 0.7 0.9 0.02 0.7 2089.0

### Mat 3
set Gs3      235000.0
set nou3     0.3
set Bs3      [expr 2.0*$Gs3*(1+$nou3)/3.0/(1-2.0*$nou3)]
set rhos3   [expr 128.0/32.2]
set cohes3  5.0
set fricAng3 37.0
set peakShearStra 0.10
set refPress 2089.0
set pressDependCoe 0.50
set PTAng3 26.0
set porosity3 0.3800
set Perm3 [expr 3.64e-4/1.938/32.2]
set alpha3 [expr (pow($xSize,2))/(4.0*($Bs3 + (4.0/3.0)*$Gs3))]
puts "$alpha3"

set bulk 41771094.402673 ; #User-Defined --- pore water bulk modulus

nDMaterial PressureDependMultiYield02 3 3 $rhos3  $Gs3 $Bs3 $fricAng3 $peakShearStra $refPress $pressDependCoe $PTAng3 0.02 0.15 0.15 0.00 20 1.50 3.0 1.0 0.0 0.61  0.9 0.02 0.7 2089.0


##########################################
#### Define Soil Elements
##########################################

set m 0
set eleNum 0

for {set i 1} {$i <= $numXele} {incr i 1} {
  for {set j 1} {$j <= $numYele} {incr j 1} {
      for {set k 1} {$k <= $numZele} {incr k 1} {
	  
          set eleNum [expr $eleNum+1]
			  
          set n1  [expr ($i-1)*($numYnode*$numZnode)+$k+($j-1)*$numZnode] 
          set n2  [expr $n1+($numYnode*$numZnode)] 
          set n3  [expr $n2+$numZnode] 
          set n4  [expr $n1+$numZnode] 
          set n5  [expr $n1+1]
          set n6  [expr $n2+1]
          set n7  [expr $n3+1]
          set n8  [expr $n4+1]  	      
			 
	 # User-Defined --- Depending on the number of soil layers, you may need more "if-else" command lines. Remember to change "$K <= 5". It should represent soil stratigraphy.
	 # At this stage, permeability must be set to 1000.0 to avoid pore water pressure generation during geostatic analysis. 	 
     if { $k <= 5 } {
          element brickUP $eleNum $n1 $n2 $n3 $n4 $n5 $n6 $n7 $n8 3  [expr $bulk/$porosity3] 1.938 1000.0 1000.0 1000.0 $gravX $gravY [expr $gravZ]
          ##element SSPbrickUP $eleNum $n1 $n2 $n3 $n4 $n5 $n6 $n7 $n8 3 [expr $bulk/$n3] 1.938 0.58 $alpha3 1000.0 1000.0 1000.0 $gravX $gravY [expr $gravZ]
        } else {
          element brickUP $eleNum $n1 $n2 $n3 $n4 $n5 $n6 $n7 $n8  2 [expr $bulk/$porosity2] 1.938 1000.0 1000.0 1000.0 $gravX $gravY [expr $gravZ]
          ##element SSPbrickUP $eleNum $n1 $n2 $n3 $n4 $n5 $n6 $n7 $n8 2 [expr $bulk/$n2] 1.938  0.7 $alpha2 1000.0 1000.0 1000.0 $gravX $gravY [expr $gravZ]
        }

    }
  }
}

##############################################
#### Prints Total Number of Nodes and Elements
##############################################
puts "Number of Nodes    =  $nodeNum"
puts "Number of Elements =  $eleNum "


##############################################
#### Boundary conditions
##############################################
# Bottom Boundary 
fixZ 0.0 1 1 1 0   

# User-defined --- Ground Water Depth With Respect to Model Base 
fixZ [expr $z0+$z1+$z2] 0 0 0 1 ; #User-Defined - Use fixZ ... 0 0 0 1 for all levels above ground water table.

# Lateral Boundaries 
 for {set k 2 } { $k <= $numZnode } { incr k 1 } {
   for {set j 2} {$j <=  $numYnode } { incr j 1 } {

       equalDOF  [expr ($numXnode-1)*($numZnode*$numYnode)+$k+(($j-2)*$numZnode)] $k  1  2 3         ;# right side yz boundary 
       equalDOF  [expr ($j-1)*$numZnode+$k] $k  1   2  3                                             ;# left  side yz boundary                                                                                
         }

       equalDOF   [expr int($numXnode*$numZnode*$numYnode-$numZnode+$k)] $k  1 2 3 

   for {set i 2} {$i <=  [expr ($numXnode-1)]} {incr i 1} {
       equalDOF  [expr ($i-1)*($numZnode*$numYnode)+$k]    $k  1   2  3                                 ;# right side xz boundary
       equalDOF  [expr $numZnode*($numYnode-1)+($i-1)*($numZnode*$numYnode)+$k]   $k   1   2 3          ;# right side xz boundary
        }
 }

puts " ## Soil is built!! ## " 


########################################################
#### Creating Output Files in a Folder titled "Outputs"
########################################################

set dataDir Outputs
file mkdir $dataDir

# Check adequacy of hydrostatic pore water pressure
eval "recorder Node    -file $dataDir/PWP_Gravity.out    -time    -nodeRange 1 $numZnode -dof 4   vel"


##############################################
#### Elastic Geostatic Analysis
##############################################

numberer Plain
system SparseSYM
#system Mumps
test NormDispIncr 1.0e-3 100 1
algorithm KrylovNewton
constraints Penalty 1.e18 1.e18
integrator Newmark 0.5 0.25
analysis Transient 

 for {set numIncr 1} {$numIncr <= 10 } {incr numIncr 1} {
       puts "##### Elastic Geostatic Analysis step : $numIncr #####";
       analyze 1 10.0
     }
	 
##############################################
#### Switch into Plastic State 
##############################################

# User-Defined --- Specify the tags of nonlinear material
for {set i 2} {$i <= 3 } {incr i 1} {
    updateMaterialStage -material $i -stage 1
   }
 
 for {set numIncr 1} {$numIncr <= 10 } {incr numIncr 1} {
       puts "##### Plastic Geostatic Analysis step : $numIncr #####";
       analyze 1 0.05
     }

##############################################
#### Preparing Model for Seismic Analysis
##############################################

wipeAnalysis
remove recorders
loadConst -time 0.0

######################################################
#### Updating Soil Permeability During Seismic Shaking
######################################################

set parameter tags 
for {set i 1} {$i <= $numZele } {incr i 1} {
    parameter [expr 100000+$i] element $i hPerm 
    parameter [expr 200000+$i] element $i vPerm 
    }

# User-Defined --- "i" values should be modified to represent soil stratigraphy.
for {set i 1} {$i <= 5 } {incr i 1} {
    updateParameter [expr 100000+$i] $Perm3 
    updateParameter [expr 200000+$i] $Perm3
    }
for {set i 6} {$i <= $numZele } {incr i 1} {
    updateParameter [expr 100000+$i] $Perm2
    updateParameter [expr 200000+$i] $Perm2
    }


##############################################
#### User-Defined Rayleigh Damping
##############################################

set zeta 0.05;		# User-Defined --- Damping Ratio in Decimal
set w1  21.12;		# User-Defined --- Angular Frequency at Frequency 1
set w2  14.96;		# User-Defined --- Angular Frequency at Frequency 2

set a0 [expr $zeta*2.0*$w1*$w2/($w1 + $w2)];	# mass damping coefficient based on first and second modes
set a1 [expr $zeta*2.0/($w1 + $w2)];			# stiffness damping coefficient based on first and second modes

rayleigh $a0 0.0 $a1 0.0; 


######################################################################
#### Removing Constraints at Model Base in the Direction of Excitation
######################################################################

for {set i 1} {$i <= $numXnode} {incr i 1} {
  for {set j 1} {$j <= $numYnode} {incr j 1} {

        remove sp [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 1  
		#remove sp [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 2 ; #Activate if motion in direction Y will be applied.
		#remove sp [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 3 ; #Activate if motion in direction Z will be applied.
      }
    } 


########################################################
#### Creating Output Files in a Folder titled "Outputs"
########################################################

eval "recorder Node    -file $dataDir/Accel_X.out   	 	-time    -nodeRange 1 $numZnode  -dof 1   accel"
eval "recorder Node    -file $dataDir/Disp_X.out   	 		-time    -nodeRange 1 $numZnode  -dof 1   disp"
eval "recorder Node    -file $dataDir/PWP_BuildUp.out  		-time    -nodeRange 1 $numZnode  -dof 4   vel"


################################################################
#### Apply Ground Motion At the Base in the Form of Displacement
################################################################

# set FACT  1
# set Gaccel 500

# timeSeries Path $Gaccel -fileTime SCTime.txt -filePath SCDisp.txt  -factor $FACT

# pattern Plain 5000 500 {

# for {set i 1} {$i <= $numXnode} {incr i 1} {
  # for {set j 1} {$j <= $numYnode} {incr j 1} {

        # sp [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 1 1.0 
		# #sp [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 2 1.0 ;# Activate if motion is applied in direction Y
		# #sp [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 3 1.0 ;# Activate if motion is applied in direction Z 		
      # }
    # }  
# }
#### Or you can use the code below to apply the motion. Both give the same results.
##pattern MultipleSupport 4000 {
##                                                                                
##groundMotion 2000 Plain -disp $Gaccel
##
##for {set i 1} {$i <= $numXnode} {incr i 1} {
##  for {set j 1} {$j <= $numYnode} {incr j 1} {
##
##        imposedMotion [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 1 2000 
##        imposedMotion [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 2 2000 ;# Activate if motion is applied in direction Y
##		 imposedMotion [expr 1+($i-1)*$numYnode*$numZnode+($j-1)*$numZnode] 3 2000 ;# Activate if motion is applied in direction Z 
##      }
##    }   
##}

########################################################################
#### End of Rigid Base Boundary Condition
########################################################################


########################################################################
#### Compliant Base Boundary Condition - Motion in the form of velocity
########################################################################

model basic -ndm 3 -ndf 3

set ID   200000
set ID_2 400000
set m    0
for {set i 1} {$i <= $numXnode} {incr i 1} {
  for {set j 1} {$j <= $numYnode} {incr j 1} {
    for {set k 1} {$k <= 1} {incr k 1} {
       set xdim [expr ($i-1)*$xSize]
       set ydim [expr ($j-1)*$ySize]

      if { $k <= [expr ($z2/$zSize2+1)] } {
          set   zdim [expr ($k-1)*$zSize2]
      } else {
         if { $k > [expr ($z2/$zSize2)+($z1/$zSize1)+1] } {
          set   zdim [expr $z2+$z1+($k-(($z2/$zSize2)+($z1/$zSize1)+1))*$zSize0]
      } else {
          set   zdim [expr $z2+($k-(($z2/$zSize2)+1))*$zSize1]
      }
      }
	   ## Top level
       node  [expr $ID+$m]    $xdim $ydim $zdim
	   ## Bottom level
	   node  [expr $ID_2+$m]  $xdim $ydim $zdim
	   
	   ## Top level
	   fix 	 [expr $ID+$m]    0 1 1 ; #User-defined --- release constraint in other directions if you want to apply motion in different directions.
	   ## Bottom level
	   fix 	 [expr $ID_2+$m]  1 1 1
	   ## Tie top level nodes to original base nodes
	   equalDOF [expr $ID+$m]  [expr $m*$numZnode+$k]   1 2 3
    }
    set m [expr ($m+1)]
  } 
}

# User-Defined --- Material properties below model base
set rockDen 25.0        ;# User-Defined  --- Rock Unit weight
set rockVs  400.0       ;# User-Defined  --- Rock shear wave velocity
set rockPoisson 0.25    ;# User-Defined  --- Rock Poisson's ratio

# Compressive wave (p-wave) --- In case you need to apply motion in vertical direction
set rockG [expr $rockVs*$rockVs*$rockDen]
set rockM [expr 2*$rockG*(1-$rockPoisson)/(1-2*$rockPoisson)]
set rockVp [expr pow($rockM/$rockDen,0.5)]


# Define Dashpot Material Elements
set m 0
for {set i 1} {$i <= $numXnode} {incr i 1} {
  for {set j 1} {$j <= $numYnode} {incr j 1} {

		uniaxialMaterial Viscous [expr 500000+$m] [expr $rockDen*$rockVs*$xSize*$ySize/4.0] 1.0
		element zeroLength       [expr 600000+$m]  [expr $ID_2+$m]  [expr $ID+$m] -mat [expr 500000+$m]   -dir  1
		set m [expr ($m+1)]
      }
    } 
	
# load pattern for u-rock 
set FACT  1
set Gaccel 1000
timeSeries Path $Gaccel -fileTime Time.txt -filePath Velocity.txt  -factor $FACT

pattern Plain 300 1000 {
	set m 0
		for {set i 1} {$i <= $numXnode} {incr i 1} {
			for {set j 1} {$j <= $numYnode} {incr j 1} {
				load    [expr $ID+$m]    [expr $rockDen*$rockVs*$xSize*$ySize/4.0] 0.0 0.0 ; # Modify if shaking is applied in other directions.
				set m [expr ($m+1)]
				}
			}

}

########################################################################
#### End of Compliant Base Boundary Condition
########################################################################


########################################
#### Define Seismic Analysis Algorithms
########################################

set compdT  0.01		; #User-Defined
set EQduration 29.50	; #User-Defined

set gamma   0.6
#integrator Newmark  $gamma  [expr pow($gamma+0.5, 2)/4.0]
integrator HHT 0.5 
#numberer RCM
numberer Plain
constraints Penalty 1.0e18 1.0e18
test NormDispIncr 1e-4 100 1
algorithm KrylovNewton
#system SparseGeneral
system SparseSYM
#system Mumps
analysis VariableTransient

 for {set numIncr 1} {$numIncr <= [expr int($EQduration/$compdT)] } {incr numIncr 1} {
        puts "##### Seismic excitation step : $numIncr #####";
        analyze 1 $compdT [expr $compdT/100.0] $compdT  100
        }

set endT [clock seconds]
set ElapsedTime [expr $endT-$startT]
set ElapsedHours [expr int($ElapsedTime/3600.)]
set ElapsedMins [expr int($ElapsedTime/60.-$ElapsedHours*60.)]
set ElapsedSecs [expr int($ElapsedTime-$ElapsedHours*3600-$ElapsedMins*60.)]
puts "Execution time: $ElapsedHours hours and $ElapsedMins minutes and $ElapsedSecs seconds !!"

########################################
#### End of the TCL Code
########################################

wipe